#
# program to turn a quake1/2/3 .map into some bsp file (format? what format?).
# all brushes of the map (of worldspawn class; don't deal with more for now)
# (note that the format is slightly different in doom3 .map files)
#
# Alex S.
#

use strict;

# check
if($#ARGV < 1){
    print "usage: [infile] [outfile]\n";
    exit;
}

print "Doing stuff... please wait.\n";
my $start = time();

print "loading map\n";
my $mfile = loadMap($ARGV[0]);
my $worldspawn = $mfile->{entsbyname}{worldspawn} or die "can't find worldspawn!\n";

print "making vertices\n";
makeverts($worldspawn);

print "list all sides\n";
my $sides = listallsides($worldspawn);

my $bspdepth=0;
print "building BSP tree\n";
my $bsp = buildBSP($sides);

print "inserting brushes (for collision detection)\n";
insertBrushes($bsp,$worldspawn);

print "writing output\n";
writeBSP($bsp,$ARGV[1]);

my $end = time();
print "Elapsed Time: ".($end-$start)." seconds.\n";

# output the produced tree, etc.
sub writeBSP {
    my ($bsp,$file) = @_;
    
    # make lists of what we want to write.
    my @nodes;   #tree nodes
    my @leafs;   #leafs 
    my $maxdepth = 1;
    my @sides;
    my @windings;
    my @brushes;

    # make a list of all the tree nodes and leafs.
    push my @stack,$bsp;
    while($#stack >= 0){
        my $node = pop @stack;
        if($node->{depth} > $maxdepth){
            $maxdepth = $node->{depth};
        }
        if($node->{type} eq 'tree'){
            push @nodes,$node;
            $node->{idx} = $#nodes;
            push @stack,$node->{front};
            push @stack,$node->{back};
        }elsif($node->{type} eq 'leaf'){
            push @leafs,$node;
            $node->{idx} = -($#leafs + 1);
        }
    }
    print "there are: ".($#nodes+1)." nodes, and ".($#leafs+1)." leafs.\n";
    print "max depth is: ".$maxdepth."\n";

    map { $_->{front} = $_->{front}{idx}; $_->{back} = $_->{back}{idx}; } @nodes;

    open my $out,'>'.$file or die $!;

    #### format ####
    # nodes
    # leafs (ignore)
    # sides
    # windings


    print $out "".($#nodes+1)."\n";
    for my $node (@nodes) {
        print $out join(' ',@{$node->{plane}}).' ';
        print $out $node->{front}.' '.$node->{back}.' ';
        print $out "".($#sides+1).' ';
        print $out "".($#{$node->{sides}}+1).' ';
        push @sides,@{$node->{sides}};
    }

    print $out "".($#leafs+1)."\n";
    for my $leaf (@leafs){
        print $out "\n".($#{$leaf->{brushes}}+1)." "; 
        print $out join(' ',@{$leaf->{brushes}})." ";
    }

    print "There are ".($#sides+1)." sides.\n";
    print $out "".($#sides+1)."\n";
    for my $side (@sides){
        print $out join(' ',@{$side->{plane}}).' ';
        print $out "".($#windings+1)." ";
        push @windings,$side->{winding};
    }
    
    print "There are ".($#windings+1)." windings.\n";
    print $out "\n".($#windings+1)."\n";
    for my $win (@windings){
        print $out "\n".($#{$win}+1)." "; 
        for my $pnt (@{$win}){
            print $out join(' ',map { sprintf("%.6f",$_); } @{$pnt})." ";
        }
    }
    
    @brushes = @{$worldspawn->{prim}};

    print "There are ".($#brushes+1)." brushes.\n";
    print $out "".($#brushes+1)."\n";
    for my $brush (@brushes){
        print $out "".($#{$brush->{sides}}+1)." ";        
        for my $side (@{$brush->{sides}}){
            print $out join(' ',@{$side->{plane}}).' ';
        }
    }
}

# add brush to tree...
sub addBrushToTree {
    my ($bsp,$b,$brushes) = @_;
    if($bsp->{type} eq 'leaf'){
        push @{$bsp->{brushes}},$b;
    }else{
        my $e = evalBrushForPlane($brushes->[$b],$bsp->{plane});
        if($e > 0){
            addBrushToTree($bsp->{front},$b,$brushes);
        }elsif($e < 0){
            addBrushToTree($bsp->{back},$b,$brushes);
        }else{
            # TODO: fix (cut the brush)
            addBrushToTree($bsp->{front},$b,$brushes);
            addBrushToTree($bsp->{back},$b,$brushes);
        }
    }
}

# insert brushes into bsp tree (for collision detection, etc.)
sub insertBrushes {
    my ($bsp,$ent) = @_;
    for my $i (0..$#{$ent->{prim}}){
        next if $ent->{prim}[$i]{type} ne 'brush';
        addBrushToTree($bsp,$i,$ent->{prim});
    }
}

# given a list of sides, pick the best one.
# this is sorta... amm... mostly guess work as to what makes the best side.
# I've tried looking for the side that splits the least other sides, but that
# takes -forever-, so I'll just pick the side whose average point is closest 
# to the overall average point.
sub pickBestSide {
    my ($sides) = @_;
    return shift @{$sides};
#
### the below doesn't seem to improve things (quite the opposite in fact)
#    my $avg = [0,0,0];
#    for my $side (@{$sides}){
#        my $t = [0,0,0];
#        map { $t = vecadd($t,$_); } @{$side->{winding}};
#        $t = vecscale($t,1.0/($#{$side->{winding}}+1));
#        $side->{avg} = $t;
#        $avg = vecadd($avg,$t);
#    }
#   $avg = vecscale($avg,1.0/($#{$sides}+1));
#
#   my $minindex = 0;
#   my $mindist = sqrt(vecdot($avg,$sides->[0]{avg}));
#   for my $i (1..$#{$sides}){
#       my $d = sqrt(vecdot($avg,$sides->[$i]{avg}));
#       if($mindist > $d){
#           $mindist = $d;
#           $minindex = $i;
#       }
#   }
#   my $side = $sides->[$minindex];
#   splice @{$sides},$minindex,1;
#   return $side;
}

sub buildBSP {
    my ($sides) = @_;
    return {type=>'leaf', depth=>$bspdepth+1, brushes=>[]} if $#{$sides} == -1;
    $bspdepth++;

    my ($f,$b) = ([],[]);
    my $node = {type=>'tree', depth=>$bspdepth };

    # pick a plane
    my $side = pickBestSide($sides);
    my $plane = $side->{plane};

    $node->{plane} = $plane;
    $node->{sides} = [ $side ];

    for my $side (@{$sides}) {
        my ($e,$in) = evalWindingForPlane($side->{winding},$plane);
        if ($e > 0) {   # front
            push @{$f},$side;
        } elsif ($e < 0) {  # back
            push @{$b},$side;
        } elsif ($e == 0 && $in == 1) {
            push @{ $node->{sides} },$side;
        } else {
            my ($front,$back) = windingchop($side->{winding},$plane);
            push @{$f},{plane=>$side->{plane},material=>$side->{material},winding=>$front};
            push @{$b},{plane=>$side->{plane},material=>$side->{material},winding=>$back};
        }
    }
    $node->{front} = buildBSP($f);
    $node->{back} = buildBSP($b);
    $bspdepth--;
    return $node;
}

# return all sides for an entity.
sub listallsides {
    my ($ent) = @_;
    my @sides;
    for my $brush (@{$ent->{prim}}){
        next if $brush->{type} ne 'brush';
        push @sides,@{$brush->{sides}};
    }
    return \@sides;
}

# given an entity (with brushes), loop for all brushes and make windings for'em.
sub makeverts {
    my ($ent) = @_;
    for my $brush (@{$ent->{prim}}){
        next if $brush->{type} ne 'brush';
        for my $side (@{$brush->{sides}}){
            $side->{winding} = baseWindingForPlane($side->{plane});
        }
        for my $i (0..$#{$brush->{sides}}) {
            for my $j (0..$#{$brush->{sides}}) {
                if($i != $j){
                    my ($f,$b) = windingchop(
                        $brush->{sides}[$j]{winding},
                        $brush->{sides}[$i]{plane}
                    );
                    $brush->{sides}[$j]{winding} = $b;
                }
            }
        }
    }
}

# load a map file (all entities, etc.).
sub loadMap {
    my ($file) = @_;
    my $cont = loadFile($file);

    # get rid of comments.
    $cont =~ s|\/\/.*?\n||sgi;

    # replace '{' and '}' with properly nested <tagN>...</tagN>, where N is the depth.
    my $cnt = 0;
    $cont =~ s/(\{|\})/$1 eq '{' ? '<tag'.(++$cnt).'>':'<\/tag'.($cnt--).'>'/sgie;

    my $ents = []; # list of ents.
    my %ents = (); # ents by name (classname)

    # loop for all entities.
    while($cont =~ m|<tag1>(.+?)</tag1>|sgi){
        my $entity = trim($1);
        my $ent = {prim=>[],vals=>{}};
        # read all of that entities brushes/patches.
        while($entity =~ m|<tag2>(.*?)</tag2>|sgi){
            my $primitive = trim($1);
            if($primitive !~ m|patch|si){  # is this a brush? (ignore patches)
                $primitive =~ s{.*<tag3>|</tag3>.*}{}sgi;
                my $brush = {type=>'brush', sides => []};
                for my $face (split /\s*\n\s*/,trim($primitive)){
                    my @pnts;
                    for(0..2){
                        push @pnts, [$1,$2,$3]
                           if $face =~ s|\(\s+(\S+)\s+(\S+)\s+(\S+)\s+\)\s+||;
                    }
                    my $material = $1 if $face =~ s|(\S+?)||;
                    my $plane = vecsplane($pnts[1],$pnts[0],$pnts[2]);
                    # ignore the rest for now.
                    push @{$brush->{sides}},{plane=>$plane,material=>$material};
                }
                push @{$ent->{prim}},$brush;
            }
        }
        # erase all of brushes/patches
        $entity =~ s|<tag2>.*</tag2>||sgi;

        # what's left is name-value pairs for that entity; read those.
        while($entity =~ m|\"(.*?)\"\s+\"(.*?)\"|sgi){
            $ent->{vals}{$1} = $2;
        }
        push @{$ents},$ent;
        $ents{$ent->{vals}{classname}} = $ent;
    }
    return {ents=>$ents,entsbyname=>\%ents,filename=>$file};
}

# load file
sub loadFile {
    my ($f) = @_;
    open my $in,$f or die $!;
    read $in,my $cont,-s $f or die $!;
    close $in;
    return $cont;
}

# cut off beginning and trailing spaces
sub trim {
    my ($s) = @_;
    $s =~ s/^\s+|\s+$//sg;
    return $s;
}

#
# Math tools for stuff...
#
# (C) 2005, Alex S.
#
# Everything is right-handed coordinate system.
# all planes are in the form: Ax + By + Cz + D = 0
#    (note that this is not the default for quake123/doom3/hl12, etc.)
# all functions that work with lists take references and return references
#    (unless otherwise noted)
#

# dot product
sub vecdot {
    my ($a,$b) = @_;
    my $sum = 0;
    map { $sum += $a->[$_] * $b->[$_] } 0..$#{$a};
    return $sum;
}

# cross product
sub veccross {
    my ($a,$b) = @_;
    return [
        $a->[1] * $b->[2] - $a->[2] * $b->[1],
        $a->[2] * $b->[0] - $a->[0] * $b->[2],
        $a->[0] * $b->[1] - $a->[1] * $b->[0]
    ];
}

# subtract vectors
sub vecsub {
    my ($a,$b) = @_;
    return [ map { $a->[$_] - $b->[$_] } 0..$#{$a} ];
}

# add vectors
sub vecadd {
    my ($a,$b) = @_;
    return [ map { $a->[$_] + $b->[$_] } 0..$#{$a} ];
}

# scale
sub vecscale {
    my ($v,$s) = @_;
    return [ map { $_ * $s } @{$v} ];
}

# normalize
sub vecnorm {
    my ($v) = @_;
    my $len = sqrt(vecdot($v,$v));
    return vecscale($v,1.0/$len);
}

# Plane is: Ax + By + Cx + D = 0
sub vecsplane {
    my ($a,$b,$c) = @_;
    my $d1 = vecsub($b,$a);
    my $d2 = vecsub($c,$a);
    my $crs = veccross($d1,$d2);
    my $norm = vecnorm($crs);
    my $d = -vecdot($a,$norm);
    return [@{$norm},$d];
}

# evaluate a point for a plane
sub evalPointForPlane {
    my ($pnt,$p) = @_;
    return vecdot($pnt,$p) + $p->[3];
}

# evaluate a point for a plane
# returns an array; first element is whether front/back/span.
# second element is whether totally on the plane or not.
sub evalWindingForPlane {
    my ($w,$p) = @_;
    my ($max,$min);
    $max = $min = evalPointForPlane($w->[0],$p);
    for my $i (1..$#{$w}){
        my $e = evalPointForPlane($w->[$i],$p);
        $max = $e if $max < $e;
        $min = $e if $min > $e;
    }
    if ($max > 0 && $min < 0) {
        return (0,1) if ($max - $min < 0.1);
        return (0,0);
    }
    return (1,0) if $max > 0;
    return (-1,0);
}

# eval brush for plane
sub evalBrushForPlane {
    my ($b,$p) = @_;
    if($#{$b->{sides}} == -1){
        return 1;
    }
    my ($max,$min) = (-1,1);
    for my $side (@{$b->{sides}}){
        my ($e,$ins) = evalWindingForPlane($side->{winding},$p);
        next if $ins;
        return 0 if $e == 0;
        $max = $e if $max < $e;
        $min = $e if $min > $e;
    }
    return 0 if $max != $min;
    return $max;
}

# makes 2 orthogonal vectors out of one.
# (we're looking down the vector; right-hand coordinate system;
# the vector is similar to +Z)
# Note: returns an arry, first parameter is 'left', another 'up'
sub vecorth {
    my ($v) = @_;
    my $d = $v->[0] * $v->[0] + $v->[1] * $v->[1];
    my $l;
    if($d <= 1.0/10000.0){
        $l = [1,0,0];
    }else{
        # the perpendicular vector is now exactly half-way
        # between projection in x,y plane, and 180 degree rotation
        # on the z plane. So rotate by 90 degrees, and normalize.
        $d = sqrt($d);
        $l = [
            -$v->[1] * 1.0 / $d,       # cos90 * x + -sin90 * y
            $v->[0] * 1.0 / $d,        # sin90 * x + cos90 * y
            0];
    }
    return ( $l,veccross($v,$l) );
}

# makes a `huge' winding that goes all the way around the polygon
sub baseWindingForPlane {
    my ($p) = @_;
    my $o = [ $p->[0] * -$p->[3],$p->[1] * -$p->[3],$p->[2] * -$p->[3] ];

    # find orthogonal left and up vectors
    my ($l,$u) = vecorth($p);

    # scale'em by some huge number.
    $l = vecscale($l,1000000000.0);
    $u = vecscale($u,1000000000.0);
    my $w = [];
    push @{$w},vecsub($o,vecsub($l,$u));
    push @{$w},vecadd($o,vecadd($l,$u));
    push @{$w},vecadd($o,vecsub($l,$u));
    push @{$w},vecsub($o,vecadd($l,$u));
    return $w;
}

# cut a 3d winding at a plane
# note: returns a list of 2 elements.
sub windingchop {
    my ($in,$pl) = @_;
    my ($f,$b) = ([],[]);
    my $s = $#{$in};
    my ($p,$k,$l,$d1,$d2,$t);
    for $p (0..$#{$in}) {
        if ( ($k = evalPointForPlane($in->[$p],$pl)) >= 0 ) {      # p is front
            if ( ($l = evalPointForPlane($in->[$s],$pl)) < 0 ) {   # s in back
                $t = vecadd($in->[$s],vecscale(vecsub($in->[$p],$in->[$s]),-$l/(-$l + $k)));
                push @{$b},$t;
                push @{$f},$t;
            }
            push @{$f},$in->[$p];
        } else {
            if ( ($l = evalPointForPlane($in->[$s],$pl)) >= 0 ) {  # s in front
                $t = vecadd($in->[$p],vecscale(vecsub($in->[$s],$in->[$p]),-$k/(-$k + $l)));
                push @{$f},$t;
                push @{$b},$t;
            }
            push @{$b},$in->[$p];
        }
        $s = $p;
    }
    return ($f,$b);
}

# some output methods
sub vec2str {
   return '('.join(', ',@{$_[0]}).')';
}
sub winding2str {
    return join(', ',map { vec2str($_) } @{$_[0]});
}
sub plane2str {
    my ($p) = @_;
    return $p->[0].'x + '.$p->[1].'y + '.$p->[2].'z + '.$p->[3].' = 0';
}


