/*
 * pVector.java
 *
 * (C) 2005, Alex S.
 */

import java.util.*;
import java.awt.*;

/**
 * basic vector routines.
 */
public class pVector {

    /**
     * given 2 points, find their sum
     */
    public static void vectorAdd (double[] va, double[] vb, double[] out){
        out[0] = va[0]+vb[0];
        out[1] = va[1]+vb[1];
        out[2] = va[2]+vb[2];
    }

    /**
     * given 2 points, find their difference
     */
    public static void vectorSubtract (double[] va, double[] vb, double[] out){
        out[0] = va[0]-vb[0];
        out[1] = va[1]-vb[1];
        out[2] = va[2]-vb[2];
    }

    /**
     * cross product of two vectors
     */
    public static void crossProduct(double[] v1, double[] v2, double[] cross ) {
        cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
        cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
        cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
    }

    /**
     * given 2 points, find their sum
     */
    public static void vectorScale (double[] v, double s, double[] out){
        out[0] = v[0]*s;
        out[1] = v[1]*s;
        out[2] = v[2]*s;
    }

    /**
     * normalize
     */
    public static double vectorNormalize(double[] in, double[] out) {
        double  length, ilength;
        length = (double)Math.sqrt(in[0]*in[0] + in[1]*in[1] + in[2]*in[2]);
        if (length == 0) {
            return 0;
        }
        ilength = (double)1.0/length;
        out[0] = in[0]*ilength;
        out[1] = in[1]*ilength;
        out[2] = in[2]*ilength;
        return length;
    }

    /**
     * dot product
     */
    public static double dotProduct (double[] v1, double[] v2){
        return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
    }

    /**
     * makes 2 orthogonal vectors out of one.
     * (we're looking down the vector; right-hand coordinate system;
     * the vector is similar to +Z)
     */
    public static void vectorOrth(double[] v, double[] left,double[] up) {
        // project x,y vector onto the z plane.
        double d = v[0] * v[0] + v[1] * v[1];
        if(d <= 1.0/10000.0) {
            // the whole vector is in Z, orthogonal `left' is just X.
            left[0] = 1; left[1] = 0; left[2] = 0;
        } else {
            // the perpendicular vector is now exactly half-way
            // between projection in x,y plane, and 180 degree rotation
            // on the z plane. So rotate by 90 degrees, and normalize.
            d = Math.sqrt(d);
            left[0] = -v[1] * 1.0/d;    // cos90 * x + -sin90 * y
            left[1] = v[0] * 1.0/d;     // sin90 * x + cos90 * y;
            left[2] = 0;
        }
        crossProduct(v,left,up);
    }
}

